@@ -1,7 +1,7 @@
 /*
  * mdadm - manage Linux "md" devices aka RAID arrays.
  *
- * Copyright (C) 2001-2002 Neil Brown <neilb@cse.unsw.edu.au>
+ * Copyright (C) 2001-2006 Neil Brown <neilb@suse.de>
  *
  *
  *    This program is free software; you can redistribute it and/or modify
@@ -28,12 +28,33 @@
  */
 
 #include	"mdadm.h"
+#include	<ctype.h>
+
+static int name_matches(char *found, char *required, char *homehost)
+{
+	/* See if the name found matches the required name, possibly
+	 * prefixed with 'homehost'
+	 */
+	char fnd[33];
+
+	strncpy(fnd, found, 32);
+	fnd[32] = 0;
+	if (strcmp(found, required)==0)
+		return 1;
+	if (homehost) {
+		int l = strlen(homehost);
+		if (l < 32 && fnd[l] == ':' &&
+		    strcmp(fnd+l+1, required)==0)
+			return 1;
+	}
+	return 0;
+}
 
 int Assemble(struct supertype *st, char *mddev, int mdfd,
-	     mddev_ident_t ident, char *conffile,
-	     mddev_dev_t devlist,
+	     mddev_ident_t ident,
+	     mddev_dev_t devlist, char *backup_file,
 	     int readonly, int runstop,
-	     char *update,
+	     char *update, char *homehost,
 	     int verbose, int force)
 {
 	/*
@@ -66,7 +87,7 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 	 *
 	 * If !uuidset and scan, look in conf-file for uuid
 	 *       If not found, give up
-	 * If !devlist and scan and uuidset, get list of devs from conf-file 
+	 * If !devlist and scan and uuidset, get list of devs from conf-file
 	 *
 	 * For each device:
 	 *   Check superblock - discard if bad
@@ -90,8 +111,9 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 	 *    START_ARRAY
 	 *
 	 */
+	int must_close = 0;
 	int old_linux = 0;
-	int vers;
+	int vers = 0; /* Keep gcc quite - it really is initialised */
 	void *first_super = NULL, *super = NULL;
 	struct {
 		char *devname;
@@ -113,34 +135,35 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 	int chosen_drive;
 	int change = 0;
 	int inargv = 0;
-	int start_partial_ok = force || devlist==NULL;
+	int start_partial_ok = (runstop >= 0) && (force || devlist==NULL || mdfd < 0);
 	unsigned int num_devs;
 	mddev_dev_t tmpdev;
 	struct mdinfo info;
-	struct mddev_ident_s ident2;
 	char *avail;
 	int nextspare = 0;
-	
-	vers = md_get_version(mdfd);
-	if (vers <= 0) {
-		fprintf(stderr, Name ": %s appears not to be an md device.\n", mddev);
-		return 1;
-	}
-	if (vers < 9000) {
-		fprintf(stderr, Name ": Assemble requires driver version 0.90.0 or later.\n"
-			"    Upgrade your kernel or try --build\n");
-		return 1;
-	}
+
 	if (get_linux_version() < 2004000)
 		old_linux = 1;
 
-	if (ioctl(mdfd, GET_ARRAY_INFO, &info.array)>=0) {
-		fprintf(stderr, Name ": device %s already active - cannot assemble it\n",
-			mddev);
-		return 1;
-	}
-	ioctl(mdfd, STOP_ARRAY, NULL); /* just incase it was started but has no content */
+	if (mdfd >= 0) {
+		vers = md_get_version(mdfd);
+		if (vers <= 0) {
+			fprintf(stderr, Name ": %s appears not to be an md device.\n", mddev);
+			return 1;
+		}
+		if (vers < 9000) {
+			fprintf(stderr, Name ": Assemble requires driver version 0.90.0 or later.\n"
+				"    Upgrade your kernel or try --build\n");
+			return 1;
+		}
 
+		if (ioctl(mdfd, GET_ARRAY_INFO, &info.array)>=0) {
+			fprintf(stderr, Name ": device %s already active - cannot assemble it\n",
+				mddev);
+			return 1;
+		}
+		ioctl(mdfd, STOP_ARRAY, NULL); /* just incase it was started but has no content */
+	}
 	/*
 	 * If any subdevs are listed, then any that don't
 	 * match ident are discarded.  Remainder must all match and
@@ -154,16 +177,20 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 	    ident->super_minor < 0 &&
 	    ident->devices == NULL) {
 		fprintf(stderr, Name ": No identity information available for %s - cannot assemble.\n",
-			mddev);
+			mddev ? mddev : "further assembly");
 		return 1;
 	}
 	if (devlist == NULL)
-		devlist = conf_get_devs(conffile);
-	else inargv = 1;
+		devlist = conf_get_devs();
+	else if (mdfd >= 0)
+		inargv = 1;
 
 	tmpdev = devlist; num_devs = 0;
 	while (tmpdev) {
-		num_devs++;
+		if (tmpdev->used)
+			tmpdev->used = 2;
+		else
+			num_devs++;
 		tmpdev = tmpdev->next;
 	}
 	devices = malloc(num_devs * sizeof(*devices));
@@ -172,16 +199,21 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 
 	if (verbose>0)
 	    fprintf(stderr, Name ": looking for devices for %s\n",
-		    mddev);
+		    mddev ? mddev : "further assembly");
 
-	while ( devlist) {
-		char *devname;
+	/* first walk the list of devices to find a consistent set
+	 * that match the criterea, if that is possible.
+	 * We flag the one we like with 'used'.
+	 */
+	for (tmpdev = devlist;
+	     tmpdev;
+	     tmpdev = tmpdev->next) {
+		char *devname = tmpdev->devname;
 		int dfd;
 		struct stat stb;
 		struct supertype *tst = st;
 
-		devname = devlist->devname;
-		devlist = devlist->next;
+		if (tmpdev->used > 1) continue;
 
 		if (ident->devices &&
 		    !match_oneof(ident->devices, devname)) {
@@ -200,22 +232,27 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 			if ((inargv && verbose >= 0) || verbose > 0)
 				fprintf(stderr, Name ": cannot open device %s: %s\n",
 					devname, strerror(errno));
+			tmpdev->used = 2;
 		} else if (fstat(dfd, &stb)< 0) {
 			/* Impossible! */
 			fprintf(stderr, Name ": fstat failed for %s: %s\n",
 				devname, strerror(errno));
+			tmpdev->used = 2;
 		} else if ((stb.st_mode & S_IFMT) != S_IFBLK) {
 			fprintf(stderr, Name ": %s is not a block device.\n",
 				devname);
+			tmpdev->used = 2;
 		} else if (!tst && (tst = guess_super(dfd)) == NULL) {
 			if ((inargv && verbose >= 0) || verbose > 0)
-				fprintf(stderr, Name ": no recogniseable superblock\n");
+				fprintf(stderr, Name ": no recogniseable superblock on %s\n",
+					devname);
+			tmpdev->used = 2;
 		} else if (tst->ss->load_super(tst,dfd, &super, NULL)) {
 			if ((inargv && verbose >= 0) || verbose > 0)
 				fprintf( stderr, Name ": no RAID superblock on %s\n",
 					 devname);
 		} else {
-			tst->ss->getinfo_super(&info, &ident2, super);
+			tst->ss->getinfo_super(&info, super);
 		}
 		if (dfd >= 0) close(dfd);
 
@@ -226,8 +263,8 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 					devname);
 			continue;
 		}
-		if (ident->name[0] &&
-		    (!super || strncmp(ident2.name, ident->name, 32)!=0)) {
+		if (ident->name[0] && (!update || strcmp(update, "name")!= 0) &&
+		    (!super || name_matches(info.name, ident->name, homehost)==0)) {
 			if ((inargv && verbose >= 0) || verbose > 0)
 				fprintf(stderr, Name ": %s has wrong name.\n",
 					devname);
@@ -254,10 +291,25 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 					devname);
 			continue;
 		}
-
-		/* If we are this far, then we are commited to this device.
+		if (mdfd < 0) {
+			if (tst == NULL || super == NULL)
+				continue;
+			if (update == NULL &&
+			    tst->ss->match_home(super, homehost)==0) {
+				if ((inargv && verbose >= 0) || verbose > 0)
+					fprintf(stderr, Name ": %s is not built for host %s.\n",
+						devname, homehost);
+				/* Auto-assemble, and this is not a usable host */
+				/* if update != NULL, we are updating the host
+				 * name... */
+				continue;
+			}
+		}
+		/* If we are this far, then we are nearly commited to this device.
 		 * If the super_block doesn't exist, or doesn't match others,
-		 * then we cannot continue
+		 * then we probably cannot continue
+		 * However if one of the arrays is for the homehost, and
+		 * the other isn't that can disambiguate.
 		 */
 
 		if (!super) {
@@ -267,9 +319,41 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 			return 1;
 		}
 
-
-		st = tst; /* commit to this format, if haven't already */
-		if (st->ss->compare_super(&first_super, super)) {
+		if (st == NULL)
+			st = tst;
+		if (st->ss != tst->ss ||
+		    st->minor_version != tst->minor_version ||
+		    st->ss->compare_super(&first_super, super) != 0) {
+			/* Some mismatch. If exactly one array matches this host,
+			 * we can resolve on that one.
+			 * Or, if we are auto assembling, we just ignore the second
+			 * for now.
+			 */
+			if (mdfd < 0)
+				continue;
+			if (homehost) {
+				int first = st->ss->match_home(first_super, homehost);
+				int last = tst->ss->match_home(super, homehost);
+				if (first+last == 1) {
+					/* We can do something */
+					if (first) {/* just ignore this one */
+						if ((inargv && verbose >= 0) || verbose > 0)
+							fprintf(stderr, Name ": %s misses out due to wrong homehost\n",
+								devname);
+						continue;
+					} else { /* reject all those sofar */
+						mddev_dev_t td;
+						if ((inargv && verbose >= 0) || verbose > 0)
+							fprintf(stderr, Name ": %s overrides previous devices due to good homehost\n",
+								devname);
+						for (td=devlist; td != tmpdev; td=td->next)
+							if (td->used == 1)
+								td->used = 0;
+						tmpdev->used = 1;
+						continue;
+					}
+				}
+			}
 			fprintf(stderr, Name ": superblock on %s doesn't match others - assembly aborted\n",
 				devname);
 			free(super);
@@ -277,12 +361,51 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 			return 1;
 		}
 
+		tmpdev->used = 1;
+	}
+
+	if (mdfd < 0) {
+		/* So... it is up to me to open the device.
+		 * We create a name '/dev/md/XXX' based on the info in the
+		 * superblock, and call open_mddev on that
+		 */
+		mdu_array_info_t inf;
+		char *c;
+		if (!first_super) {
+			return 2;
+		}
+		st->ss->getinfo_super(&info, first_super);
+		c = strchr(info.name, ':');
+		if (c) c++; else c= info.name;
+		if (isdigit(*c) && ((ident->autof & 7)==4 || (ident->autof&7)==6))
+			/* /dev/md/d0 style for partitionable */
+			asprintf(&mddev, "/dev/md/d%s", c);
+		else
+			asprintf(&mddev, "/dev/md/%s", c);
+		mdfd = open_mddev(mddev, ident->autof);
+		if (mdfd < 0)
+			return mdfd;
+		vers = md_get_version(mdfd);
+		if (ioctl(mdfd, GET_ARRAY_INFO, &inf)==0) {
+			fprintf(stderr, Name ": %s already active, cannot restart it!\n", mddev);
+			close(mdfd);
+			free(first_super);
+			return 1;
+		}
+		must_close = 1;
+	}
+
+	/* Ok, no bad inconsistancy, we can try updating etc */
+	for (tmpdev = devlist; tmpdev; tmpdev=tmpdev->next) if (tmpdev->used == 1) {
+		char *devname = tmpdev->devname;
+		struct stat stb;
 		/* looks like a good enough match to update the super block if needed */
+#ifndef MDASSEMBLE
 		if (update) {
+			int dfd;
 			/* prepare useful information in info structures */
 			struct stat stb2;
 			fstat(mdfd, &stb2);
-			info.array.md_minor = minor(stb2.st_rdev);
 
 			if (strcmp(update, "uuid")==0 &&
 			    !ident->uuid_set) {
@@ -295,13 +418,29 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 					*(__u32*)(ident->uuid+3) = random();
 				}
 				if (rfd >= 0) close(rfd);
-				ident->uuid_set = 1;
 			}
-			memcpy(info.uuid, ident->uuid, 16);
-			st->ss->update_super(&info, super, update, devname, verbose);
-			
 			dfd = dev_open(devname, O_RDWR|O_EXCL);
-			if (dfd < 0) 
+
+			if (super) {
+				free(super);
+				super = NULL;
+			}
+
+			st->ss->load_super(st, dfd, &super, NULL);
+			st->ss->getinfo_super(&info, super);
+
+			memcpy(info.uuid, ident->uuid, 16);
+			strcpy(info.name, ident->name);
+			info.array.md_minor = minor(stb2.st_rdev);
+
+			st->ss->update_super(&info, super, update, devname, verbose,
+					     ident->uuid_set, homehost);
+			if (strcmp(update, "uuid")==0 &&
+			    !ident->uuid_set) {
+				ident->uuid_set = 1;
+				memcpy(ident->uuid, info.uuid, 16);
+			}
+			if (dfd < 0)
 				fprintf(stderr, Name ": Cannot open %s for superblock update\n",
 					devname);
 			else if (st->ss->store_super(st, dfd, super))
@@ -309,8 +448,30 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 					devname);
 			if (dfd >= 0)
 				close(dfd);
+
+			if (strcmp(update, "uuid")==0 &&
+			    ident->bitmap_fd)
+				if (bitmap_update_uuid(ident->bitmap_fd, info.uuid) != 0)
+					fprintf(stderr, Name ": Could not update uuid on %s.\n",
+						devname);
+		} else
+#endif
+		{
+			int dfd;
+			dfd = dev_open(devname, O_RDWR|O_EXCL);
+
+			if (super) {
+				free(super);
+				super = NULL;
+			}
+
+			st->ss->load_super(st, dfd, &super, NULL);
+			st->ss->getinfo_super(&info, super);
+			close(dfd);
 		}
 
+		stat(devname, &stb);
+
 		if (verbose > 0)
 			fprintf(stderr, Name ": %s is identified as a member of %s, slot %d.\n",
 				devname, mddev, info.disk.raid_disk);
@@ -329,7 +490,7 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 			    > devices[most_recent].events)
 				most_recent = devcnt;
 		}
-		if (info.array.level == -4) 
+		if (info.array.level == -4)
 			/* with multipath, the raid_disk from the superblock is meaningless */
 			i = devcnt;
 		else
@@ -338,6 +499,10 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 			if (nextspare < info.array.raid_disks)
 				nextspare = info.array.raid_disks;
 			i = nextspare++;
+		} else {
+			if (i >= info.array.raid_disks &&
+			    i >= nextspare)
+				nextspare = i+1;
 		}
 		if (i < 10000) {
 			if (i >= bestcnt) {
@@ -353,6 +518,25 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 				best = newbest;
 				bestcnt = newbestcnt;
 			}
+			if (best[i] >=0 &&
+			    devices[best[i]].events == devices[devcnt].events &&
+			    devices[best[i]].minor != devices[devcnt].minor &&
+			    st->ss->major == 0 &&
+			    info.array.level != -4) {
+				/* two different devices with identical superblock.
+				 * Could be a mis-detection caused by overlapping
+				 * partitions.  fail-safe.
+				 */
+				fprintf(stderr, Name ": WARNING %s and %s appear"
+					" to have very similar superblocks.\n"
+					"      If they are really different, "
+					"please --zero the superblock on one\n"
+					"      If they are the same, please remove "
+					"one from the list.\n",
+					devices[best[i]].devname, devname);
+				if (must_close) close(mdfd);
+				return 1;
+			}
 			if (best[i] == -1
 			    || devices[best[i]].events < devices[devcnt].events)
 				best[i] = devcnt;
@@ -371,10 +555,11 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 		fprintf(stderr, Name ": no devices found for %s\n",
 			mddev);
 		free(first_super);
+		if (must_close) close(mdfd);
 		return 1;
 	}
 
-	st->ss->getinfo_super(&info, &ident2, first_super);
+	st->ss->getinfo_super(&info, first_super);
 
 	/* now we have some devices that might be suitable.
 	 * I wonder how many
@@ -448,7 +633,7 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 			continue;
 		}
 		info.events = devices[most_recent].events;
-		st->ss->update_super(&info, super, "force", devices[chosen_drive].devname, verbose);
+		st->ss->update_super(&info, super, "force", devices[chosen_drive].devname, verbose, 0, NULL);
 
 		if (st->ss->store_super(st, fd, super)) {
 			close(fd);
@@ -486,21 +671,24 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 		if ((fd=dev_open(devices[j].devname, O_RDONLY|O_EXCL))< 0) {
 			fprintf(stderr, Name ": Cannot open %s: %s\n",
 				devices[j].devname, strerror(errno));
+			if (must_close) close(mdfd);
 			return 1;
 		}
 		if (st->ss->load_super(st,fd, &super, NULL)) {
 			close(fd);
 			fprintf(stderr, Name ": RAID superblock has disappeared from %s\n",
 				devices[j].devname);
+			if (must_close) close(mdfd);
 			return 1;
 		}
 		close(fd);
 	}
 	if (super == NULL) {
 		fprintf(stderr, Name ": No suitable drives found for %s\n", mddev);
+		if (must_close) close(mdfd);
 		return 1;
 	}
-	st->ss->getinfo_super(&info, &ident2, super);
+	st->ss->getinfo_super(&info, super);
 	for (i=0; i<bestcnt; i++) {
 		int j = best[i];
 		unsigned int desired_state;
@@ -519,7 +707,7 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 		info.disk.state = desired_state;
 
 		if (devices[j].uptodate &&
-		    st->ss->update_super(&info, super, "assemble", NULL, verbose)) {
+		    st->ss->update_super(&info, super, "assemble", NULL, verbose, 0, NULL)) {
 			if (force) {
 				if (verbose >= 0)
 					fprintf(stderr, Name ": "
@@ -543,27 +731,63 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 	}
 	if (force && okcnt == info.array.raid_disks-1) {
 		/* FIXME check event count */
-		change += st->ss->update_super(&info, super, "force", 
-					devices[chosen_drive].devname, verbose);
+		change += st->ss->update_super(&info, super, "force",
+					devices[chosen_drive].devname, verbose, 0, NULL);
 	}
 
 	if (change) {
 		int fd;
 		fd = dev_open(devices[chosen_drive].devname, O_RDWR|O_EXCL);
 		if (fd < 0) {
-			fprintf(stderr, Name ": Could open %s for write - cannot Assemble array.\n",
+			fprintf(stderr, Name ": Could not open %s for write - cannot Assemble array.\n",
 				devices[chosen_drive].devname);
+			if (must_close) close(mdfd);
 			return 1;
 		}
 		if (st->ss->store_super(st, fd, super)) {
 			close(fd);
 			fprintf(stderr, Name ": Could not re-write superblock on %s\n",
 				devices[chosen_drive].devname);
+			if (must_close) close(mdfd);
 			return 1;
 		}
 		close(fd);
 	}
 
+	/* If we are in the middle of a reshape we may need to restore saved data
+	 * that was moved aside due to the reshape overwriting live data
+	 * The code of doing this lives in Grow.c
+	 */
+#ifndef MDASSEMBLE
+	if (info.reshape_active) {
+		int err = 0;
+		int *fdlist = malloc(sizeof(int)* bestcnt);
+		for (i=0; i<bestcnt; i++) {
+			int j = best[i];
+			if (j >= 0) {
+				fdlist[i] = dev_open(devices[j].devname, O_RDWR|O_EXCL);
+				if (fdlist[i] < 0) {
+					fprintf(stderr, Name ": Could not open %s for write - cannot Assemble array.\n",
+						devices[j].devname);
+					err = 1;
+					break;
+				}
+			} else
+				fdlist[i] = -1;
+		}
+		if (!err)
+			err = Grow_restart(st, &info, fdlist, bestcnt, backup_file);
+		while (i>0) {
+			i--;
+			if (fdlist[i]>=0) close(fdlist[i]);
+		}
+		if (err) {
+			fprintf(stderr, Name ": Failed to restore critical section for reshape, sorry.\n");
+			if (must_close) close(mdfd);
+			return err;
+		}
+	}
+#endif
 	/* count number of in-sync devices according to the superblock.
 	 * We must have this number to start the array without -s or -R
 	 */
@@ -578,21 +802,39 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 			inf.major_version = st->ss->major;
 			inf.minor_version = st->minor_version;
 			rv = ioctl(mdfd, SET_ARRAY_INFO, &inf);
-		} else 
+		} else
 			rv = ioctl(mdfd, SET_ARRAY_INFO, NULL);
 
 		if (rv) {
 			fprintf(stderr, Name ": SET_ARRAY_INFO failed for %s: %s\n",
 				mddev, strerror(errno));
+			if (must_close) close(mdfd);
 			return 1;
 		}
 		if (ident->bitmap_fd >= 0) {
 			if (ioctl(mdfd, SET_BITMAP_FILE, ident->bitmap_fd) != 0) {
 				fprintf(stderr, Name ": SET_BITMAP_FILE failed.\n");
+				if (must_close) close(mdfd);
 				return 1;
 			}
+		} else if (ident->bitmap_file) {
+			/* From config file */
+			int bmfd = open(ident->bitmap_file, O_RDWR);
+			if (bmfd < 0) {
+				fprintf(stderr, Name ": Could not open bitmap file %s\n",
+					ident->bitmap_file);
+				if (must_close) close(mdfd);
+				return 1;
+			}
+			if (ioctl(mdfd, SET_BITMAP_FILE, bmfd) != 0) {
+				fprintf(stderr, Name ": Failed to set bitmapfile for %s\n", mddev);
+				close(bmfd);
+				if (must_close) close(mdfd);
+				return 1;
+			}
+			close(bmfd);
 		}
-					
+
 		/* First, add the raid disks, but add the chosen one last */
 		for (i=0; i<= bestcnt; i++) {
 			int j;
@@ -626,7 +868,7 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 		}
 		
 		if (runstop == 1 ||
-		    (runstop == 0 && 
+		    (runstop <= 0 &&
 		     ( enough(info.array.level, info.array.raid_disks, info.array.layout, avail, okcnt) &&
 		       (okcnt >= req_cnt || start_partial_ok)
 			     ))) {
@@ -634,21 +876,52 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 				if (verbose >= 0) {
 					fprintf(stderr, Name ": %s has been started with %d drive%s",
 						mddev, okcnt, okcnt==1?"":"s");
-					if (okcnt < info.array.raid_disks) 
+					if (okcnt < info.array.raid_disks)
 						fprintf(stderr, " (out of %d)", info.array.raid_disks);
 					if (sparecnt)
 						fprintf(stderr, " and %d spare%s", sparecnt, sparecnt==1?"":"s");
 					fprintf(stderr, ".\n");
 				}
+				if (must_close) {
+					int usecs = 1;
+					close(mdfd);
+					/* There is a nasty race with 'mdadm --monitor'.
+					 * If it opens this device before we close it,
+					 * it gets an incomplete open on which IO
+					 * doesn't work and the capacity if wrong.
+					 * If we reopen (to check for layered devices)
+					 * before --monitor closes, we loose.
+					 *
+					 * So: wait upto 1 second for there to be
+					 * a non-zero capacity.
+					 */
+					while (usecs < 1000) {
+						mdfd = open(mddev, O_RDONLY);
+						if (mdfd >= 0) {
+							unsigned long size;
+							if (ioctl(mdfd, BLKGETSIZE, &size) == 0 &&
+							    size > 0)
+								break;
+							close(mdfd);
+						}
+						usleep(usecs);
+						usecs <<= 1;
+					}
+				}
 				return 0;
 			}
 			fprintf(stderr, Name ": failed to RUN_ARRAY %s: %s\n",
 				mddev, strerror(errno));
+			if (must_close) close(mdfd);
 			return 1;
 		}
 		if (runstop == -1) {
-			fprintf(stderr, Name ": %s assembled from %d drive%s, but not started.\n",
+			fprintf(stderr, Name ": %s assembled from %d drive%s",
 				mddev, okcnt, okcnt==1?"":"s");
+			if (okcnt != info.array.raid_disks)
+				fprintf(stderr, " (out of %d)", info.array.raid_disks);
+			fprintf(stderr, ", but not started.\n");
+			if (must_close) close(mdfd);
 			return 0;
 		}
 		if (verbose >= 0) {
@@ -665,6 +938,7 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 				fprintf(stderr, " (use --run to insist).\n");
 			}
 		}
+		if (must_close) close(mdfd);
 		return 1;
 	} else {
 		/* The "chosen_drive" is a good choice, and if necessary, the superblock has
@@ -680,5 +954,6 @@ int Assemble(struct supertype *st, char *mddev, int mdfd,
 		}
 		
 	}
+	if (must_close) close(mdfd);
 	return 0;
 }
